package testy

import (
	"io/ioutil"
	"net/http"
	"net/http/httptest"
	"net/url"
	"regexp"
	"strings"
	"testing"
)

func TestHTTPRequest(t *testing.T) {
	type tt struct {
		expected, actual interface{}
		re               []Replacement
		result           string
	}

	tests := NewTable()
	tests.Add("two GET requests", tt{
		expected: httptest.NewRequest("GET", "/", nil),
		actual:   httptest.NewRequest("GET", "/", nil),
	})
	tests.Add("different URLs requests", tt{
		expected: httptest.NewRequest("GET", "/", nil),
		actual:   httptest.NewRequest("GET", "/foo.html", nil),
		result:   "--- expected\n+++ actual\n@@ -1,3 +1,3 @@\n-GET / HTTP/1.1\r\n+GET /foo.html HTTP/1.1\r\n Host: example.com\r\n \r\n",
	})
	tests.Add("nil request", tt{
		expected: nil,
		actual:   httptest.NewRequest("GET", "/foo.html", nil),
		result:   "--- expected\n+++ actual\n@@ -1 +1,3 @@\n-\n+GET /foo.html HTTP/1.1\r\n+Host: example.com\r\n+\r\n",
	})
	tests.Add("string", tt{
		expected: `GET / HTTP/1.1
Host: localhost:6005
User-Agent: curl/7.52.1
Accept: */*

`,
		actual: &http.Request{
			Method:     http.MethodGet,
			ProtoMajor: 1,
			ProtoMinor: 1,
			URL:        &url.URL{Host: "localhost:6005"},
			Header: http.Header{
				"Accept":     []string{"*/*"},
				"User-Agent": []string{"curl/7.52.1"},
			},
		},
	})
	tests.Add("byte slice", tt{
		expected: []byte(`GET / HTTP/1.1
Host: localhost:6005
User-Agent: curl/7.52.1
Accept: */*

`),
		actual: &http.Request{
			Method:     http.MethodGet,
			ProtoMajor: 1,
			ProtoMinor: 1,
			URL:        &url.URL{Host: "localhost:6005"},
			Header: http.Header{
				"Accept":     []string{"*/*"},
				"User-Agent": []string{"curl/7.52.1"},
			},
		},
	})
	tests.Add("replacement", tt{
		expected: []byte(`GET / HTTP/1.1
Host: tsohlacol:6005
User-Agent: curl/7.52.1
Accept: */*

`),
		actual: &http.Request{
			Method:     http.MethodGet,
			ProtoMajor: 1,
			ProtoMinor: 1,
			URL:        &url.URL{Host: "localhost:6005"},
			Header: http.Header{
				"Accept":     []string{"*/*"},
				"User-Agent": []string{"curl/7.52.1"},
			},
		},
		re: []Replacement{
			{
				Regexp:      regexp.MustCompile("localhost"),
				Replacement: "tsohlacol",
			},
		},
	})
	tests.Add("file", tt{
		expected: &File{Path: "testdata/request.raw"},
		actual: &http.Request{
			Method:     http.MethodGet,
			ProtoMajor: 1,
			ProtoMinor: 1,
			URL:        &url.URL{Host: "localhost:6005"},
			Header: http.Header{
				"Accept":     []string{"*/*"},
				"User-Agent": []string{"curl/7.52.1"},
			},
		},
	})
	tests.Add("nil", tt{
		expected: nil,
		actual:   nil,
	})
	tests.Add("unknown input type", tt{expected: 123,
		actual: nil,
		result: "Failed to dump expected request: Unable to convert int to *http.Request",
	})
	tests.Add("nil body", func(t *testing.T) interface{} {
		req := httptest.NewRequest("PUT", "/foo", nil)
		req.Body = nil

		return tt{
			expected: Snapshot(t),
			actual:   req,
		}
	})

	tests.Run(t, func(t *testing.T, tt tt) {
		result := DiffHTTPRequest(tt.expected, tt.actual, tt.re...)
		var resultText string
		if result != nil {
			resultText = result.String()
		}
		if d := DiffText(tt.result, resultText); d != nil {
			t.Error(d)
		}
	})
}

func TestHTTPResponse(t *testing.T) {
	tests := []struct {
		name             string
		expected, actual interface{}
		re               []Replacement
		result           string
	}{
		{
			name:     "two empty responses",
			expected: &http.Response{},
			actual:   &http.Response{},
		},
		{
			name: "Different headers",
			expected: &http.Response{
				Header: http.Header{"Foo": []string{"bar"}},
			},
			actual: &http.Response{
				Header: http.Header{"Foo": []string{"qux"}},
			},
			result: "--- expected\n+++ actual\n@@ -1,4 +1,4 @@\n HTTP/0.0 000 status code 0\r\n-Foo: bar\r\n+Foo: qux\r\n Content-Length: 0\r\n \r\n",
		},
		{
			name:     "nil response body",
			expected: nil,
			actual: &http.Response{
				Header: http.Header{"Foo": []string{"qux"}},
			},
			result: "--- expected\n+++ actual\n@@ -1 +1,4 @@\n-\n+HTTP/0.0 000 status code 0\r\n+Foo: qux\r\n+Content-Length: 0\r\n+\r\n",
		},
		{
			name: "two nil bodies",
			expected: &http.Response{
				StatusCode: 400,
				ProtoMajor: 1,
				ProtoMinor: 1,
				Header:     http.Header{"Foo": []string{"qux"}},
			},
			actual: &http.Response{
				StatusCode: 400,
				ProtoMajor: 1,
				ProtoMinor: 1,
				Header:     http.Header{"Foo": []string{"qux"}},
			},
		},
		{
			name:     "read from file",
			expected: &File{Path: "testdata/response.raw"},
			actual: &http.Response{
				StatusCode:    http.StatusOK,
				ProtoMajor:    1,
				ProtoMinor:    1,
				ContentLength: 11,
				Header: http.Header{
					"Content-Type": []string{"application/json"},
					"Date":         []string{"Tue, 22 Jan 2019 18:44:09 GMT"},
				},
				Body: ioutil.NopCloser(strings.NewReader(`{"ok":true}`)),
			},
		},
		{
			name: "string",
			expected: `HTTP/1.1 200 OK
Content-Length: 11
Content-Type: application/json
Date: Tue, 22 Jan 2019 18:44:09 GMT

{"ok":true}
`,
			actual: &http.Response{
				StatusCode:    http.StatusOK,
				ProtoMajor:    1,
				ProtoMinor:    1,
				ContentLength: 11,
				Header: http.Header{
					"Content-Type": []string{"application/json"},
					"Date":         []string{"Tue, 22 Jan 2019 18:44:09 GMT"},
				},
				Body: ioutil.NopCloser(strings.NewReader(`{"ok":true}`)),
			},
		},
		{
			name: "byte slice",
			expected: []byte(`HTTP/1.1 200 OK
Content-Length: 11
Content-Type: application/json
Date: Tue, 22 Jan 2019 18:44:09 GMT

{"ok":true}
`),
			actual: &http.Response{
				StatusCode:    http.StatusOK,
				ProtoMajor:    1,
				ProtoMinor:    1,
				ContentLength: 11,
				Header: http.Header{
					"Content-Type": []string{"application/json"},
					"Date":         []string{"Tue, 22 Jan 2019 18:44:09 GMT"},
				},
				Body: ioutil.NopCloser(strings.NewReader(`{"ok":true}`)),
			},
		},
		{
			name: "replacement",
			expected: []byte(`HTTP/1.1 200 OK
Content-Length: 11
Content-Type: application/nosj
Date: Tue, 22 Jan 2019 18:44:09 GMT

{"ok":true}
`),
			actual: &http.Response{
				StatusCode:    http.StatusOK,
				ProtoMajor:    1,
				ProtoMinor:    1,
				ContentLength: 11,
				Header: http.Header{
					"Content-Type": []string{"application/json"},
					"Date":         []string{"Tue, 22 Jan 2019 18:44:09 GMT"},
				},
				Body: ioutil.NopCloser(strings.NewReader(`{"ok":true}`)),
			},
			re: []Replacement{
				{
					Regexp:      regexp.MustCompile(`json`),
					Replacement: `nosj`,
				},
			},
		},
		{
			name:     "unknown input type",
			expected: int(123),
			result:   "Failed to dump expected response: Unable to convert int to *http.Response",
		},
	}
	for _, test := range tests {
		t.Run(test.name, func(t *testing.T) {
			result := DiffHTTPResponse(test.expected, test.actual, test.re...)
			var resultText string
			if result != nil {
				resultText = result.String()
			}
			if d := DiffText(test.result, resultText); d != nil {
				t.Error(d)
			}
		})
	}
}

func Test_dumpRequest(t *testing.T) {
	type tt struct {
		req interface{}
		err string
	}

	tests := NewTable()
	tests.Add("invalid request", tt{
		req: 123,
		err: "Unable to convert int to *http.Request",
	})
	tests.Add("do not add Content-Length", tt{
		req: httptest.NewRequest(http.MethodPost, "/", strings.NewReader("foo")),
	})
	tests.Add("add Content-Length", tt{
		req: httptest.NewRequest(http.MethodGet, "/", nil),
	})
	tests.Add("add 0 Content-Length", tt{
		req: httptest.NewRequest(http.MethodPost, "/", nil),
	})

	tests.Run(t, func(t *testing.T, tt tt) {
		got, err := dumpRequest(tt.req)
		Error(t, tt.err, err)
		if d := DiffInterface(Snapshot(t), got); d != nil {
			t.Error(d)
		}
	})
}
